/*
FILE:     DictObjCont.C
*/
/*
VERSION:  8.004
*/
/*
DATE:     5/10/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


/*!
** \file DictObjCont.C
**
** \brief Implementation file for ObjCont, ItemObjCont and DictObjCont classes.
*/


#include <string>

#include "ISTable.h"
#include "DDLParserBase.h"
#include "DICParserBase.h"
#include "DictObjContInfo.h"
#include "DictObjCont.h"


ObjCont::ObjCont(FileNavigator& fileNav, DicFile& dicFile,
  const string& blockName, const string& id, const ObjContInfo& objContInfo) :
  _fileNav(fileNav), _dicFile(dicFile), _objContInfo(objContInfo)
{
  Init();

#ifdef VLAD_ADD_THIS
  // Throw exception here and everywhere regarding blockName or id or
  // maybe something else.
  if (_blockName.empty() || _id.empty())
      return KEYS_NOT_SET; 
#endif

  _blockName = blockName;
  _id = id;  

  _verbose = false;
}


ObjCont::~ObjCont()
{
  _blockName.clear();
  _id.clear();

  _index.clear();

  _itemsStore.clear();

}


void ObjCont::Init()
{
  _blockName.clear();
  _id.clear();

  _index.clear();

  _itemsStore.clear();

  vector<string> tmp1;
  vector<vector<string> > tmp2;

  for (unsigned int i = 0; i < _objContInfo._cats.size(); ++i)
  {
    tmp2.clear();
    for (unsigned int j = 0; j < _objContInfo._cats[i].items.size();
      ++j)
    {
      tmp2.push_back(tmp1);
    }
    _itemsStore.push_back(tmp2);
  }

}


string ObjCont::GetName()
{

    return(_id);

}


const vector<string>& ObjCont::GetAttribute(const string& catName,
  const string& itemName) const
{

    pair<unsigned int, unsigned int> indexPair =
      _objContInfo.GetItemIndices(catName, itemName);

    return(_itemsStore[indexPair.first][indexPair.second]);

}


void ObjCont::Print()
{
    cout << endl;
    cout << "Dictionary/Datablock Name: " << _blockName << endl;
    cout << "Id: " << _id << endl;

    pair<unsigned int, unsigned int> indexPair;

    for (unsigned int i = 0; i < _objContInfo._cats.size(); ++i)
    {
        for (unsigned int j = 0; j < _objContInfo._cats[i].\
          items.size(); ++j)
        {
            indexPair = _objContInfo.GetItemIndices(_objContInfo.\
              _cats[i].catName.c_str(),
            _objContInfo._cats[i].items[j].itemName.c_str());
            cout << _objContInfo._cats[i].items[j].descr << ": (" <<
              _itemsStore[indexPair.first][indexPair.second].size() << ")" << endl;

            for (unsigned int ii = 0; ii < _itemsStore[indexPair.first]\
              [indexPair.second].size(); ii++) 
                cout << "  #" << ii << " : " << _itemsStore[indexPair.first]\
                  [indexPair.second][ii] << endl;
        }
    }

}


int ObjCont::Write()
{
    // Write items
    for (unsigned int i = 0; i < _itemsStore.size(); ++i)
    {
        for (unsigned int j = 0; j < _itemsStore[i].size(); ++j)
        {
            _index.push_back(WriteItem(make_pair(i, j)));
        }
    }

    Word i = 0;
    int err = _fileNav.WriteWords(_index, i);
    if (err == NO_ERROR) 
        return i; 
    else
        return err;
}


int ObjCont::WriteItem(const pair<unsigned int, unsigned int>& indexPair)
{
    Word i = 0;
    int err;

    if (_itemsStore[indexPair.first][indexPair.second].empty())
        return NO_ERROR;

    if ((err = _fileNav.WriteStrings(_itemsStore[indexPair.first]\
      [indexPair.second], i)) == NO_ERROR)
        return i;
    else
        return err;
}


int ObjCont::Read(uWord which, unsigned int Index)
{
/*
-------------------------------------------------------------------------------
  Read() calls all of the private Get methods so that the entire object
  can be retrieved from memory. The object# must be specified.
-------------------------------------------------------------------------------
*/
  int err;
  err = 0;

  _index.clear();
  _fileNav.GetWords(which, _index, err); 

  pair<unsigned int, unsigned int> indexPair;

  // Get complex items
  for (unsigned int i = 0;
    i < _objContInfo._cats.size(); ++i)
  {
    for (unsigned int j = 0; j < _objContInfo._cats[i].items.size(); ++j)
    {
      indexPair = _objContInfo.GetItemIndices(_objContInfo._cats[i].catName.c_str(),
        _objContInfo._cats[i].items[j].itemName.c_str());

      ReadItem(indexPair, Index);
      ++Index;
    }
  }

  if (err == NO_ERROR) 
    return NO_ERROR; 
  else return err;
}


int ObjCont::ReadItem(const pair<unsigned int, unsigned int>& indexPair,
  unsigned int Index)
{
    int err;
    err = 0;
    if (_index.empty()) return INDEX_NOT_CONSTRUCTED;

  
    if (!_itemsStore[indexPair.first][indexPair.second].empty())
    {
        for (uWord i = 0; i < _itemsStore[indexPair.first][indexPair.second].\
          size(); i++) 
          _itemsStore[indexPair.first][indexPair.second][i].clear();
        _itemsStore[indexPair.first][indexPair.second].clear();
    }

    if (_index[Index] == 0)
        _itemsStore[indexPair.first][indexPair.second].clear();
    else
        _fileNav.GetStrings(_itemsStore[indexPair.first][indexPair.second],
          _index[Index], err);
    return err; 
}


void ObjCont::SetVerbose(bool verbose)
{
    _verbose = verbose;
}


void ObjCont::Build()
{

    for (unsigned int i = 0; i < _objContInfo._cats.size(); ++i)
    {
        BuildItems(_itemsStore[i], i);
    }

}


void ObjCont::BuildItems(vector<vector<string> >& combo,
  const unsigned int configIndex)
{
    ObjCont::BuildItems(combo, configIndex, _id);    
}


void ObjCont::BuildItems(vector<vector<string> >& combo,
  const unsigned int configIndex, const string& value)
{
    int i;
    ISTable * tbl;
    vector<string> ColumnNames;
    vector<string> values;
    vector<unsigned int> listOut;
    string cell;
    vector<string> columnNames;

    if (_id.empty() || _blockName.empty())
    {
        return;
    }
  
    Block& block = _dicFile.GetBlock(_blockName);

    tbl = block.GetTablePtr(_objContInfo._cats[configIndex].catName);
    if (tbl == NULL)
    {
        return;
    }

    ColumnNames.push_back(_objContInfo._cats[configIndex].col1);
    if (_objContInfo._cats[configIndex].nonDefaultValue)
    {
        unsigned int valIndex =
          _objContInfo.GetItemIndex(ObjContInfo::CIF_DDL_CATEGORY_ITEM_TYPE,
          ObjContInfo::CIF_DDL_ITEM_CODE);

        vector<string> tmp1;
        vector<vector<string> > valueVector;

        for (unsigned int j = 0; j < _objContInfo._cats[valIndex].\
          items.size(); ++j)
        {
            valueVector.push_back(tmp1);
        }

        BuildItems(valueVector, valIndex);

        if (valueVector[0].empty())
            return;
        values.push_back(valueVector[0][0]); 

        if (_itemsStore[valIndex][0].empty())
            _itemsStore[valIndex][0].push_back(valueVector[0][0]);
    }
    else
    {
        values.push_back(value); 
    }

    tbl->SetFlags(_objContInfo._cats[configIndex].col1, ISTable::DT_STRING |
      ISTable::CASE_INSENSE);

    tbl->Search(listOut, values, ColumnNames);
    if (listOut.empty())
    {
        return;
    }

    //try
    //{
    for (unsigned int ind = 0; ind < _objContInfo.\
      _cats[configIndex].items.size(); ++ind)
    {
        if (tbl->IsColumnPresent(_objContInfo._cats[configIndex].\
          items[ind].itemName))
        {
            columnNames.push_back(_objContInfo._cats[configIndex].\
              items[ind].itemName);
        }
        else
        {
            // VLAD - CODE ENHANCEMENTS - FIX THIS
            columnNames.push_back(String::Empty);
        }
    }

    for (i = 0; i < (int)listOut.size(); i++)
    {
        for (unsigned int ind = 0; ind < _objContInfo.\
          _cats[configIndex].items.size(); ++ind)
        {
            // VLAD - CODE ENHANCEMENTS - FIX THIS
            if (!columnNames[ind].empty())
            {
                //tbl->GetCell(cell, columnNames[ind], listOut[i]);
                cell = (*tbl)(listOut[i], columnNames[ind]);
            }
            combo[ind].push_back(cell);
            cell.clear();
        }
    }
    //}
    //catch (...)
    //{
      // VLAD - BUG - just continue execution. address this later
      // Reproduce in dict suite: make odb
    //}


}


ItemObjCont::ItemObjCont(FileNavigator& fileNav, DicFile& dicFile,
  const string& blockName, const string& itemName) :
  ObjCont(fileNav, dicFile, blockName, itemName, RcsbItem)
{
    GetItemDecendency(_decendency);
}


ItemObjCont::~ItemObjCont()
{
    _decendency.clear();
}


void ItemObjCont::Build()
{

    for (unsigned int i = 0; i < _objContInfo._cats.size(); ++i)
    {
        if (_objContInfo._cats[i].inheritance)
            BuildItems(_itemsStore[i], i);
        else
            ObjCont::BuildItems(_itemsStore[i], i);
    }

}


void ItemObjCont::GetItemDecendency(vector<family_member>& decendency)
/* -------------------------------------------------------------------------
   GetItemDecendency(): creates and returns a complete Family for the item
   pointed to by blockName, categoryName, and itemKeyword.  The number of
   decendents in the Family is returned in the reference numD.
   ------------------------------------------------------------------------- */
{
    vector<string> ColumnNamesCh;
    vector<string> ColumnNamesPr;
    vector<string> values;
    vector<unsigned int> listOut;
    string cell;
    int num,len, num1;

    decendency.clear();

    int * parentList, i;
    int pCount,totalCount, generationNum;
    pCount = 0;
    totalCount = 0;
    generationNum = 1;
    if (_blockName.empty()) return;
  
    Block& block = _dicFile.GetBlock(_blockName);

    ISTable *tbl = block.GetTablePtr(ObjContInfo::CIF_DDL_CATEGORY_ITEM_LINKED);
    if(tbl == NULL) return;
    ColumnNamesCh.push_back(ObjContInfo::CIF_DDL_ITEM_CHILD_NAME);
    ColumnNamesPr.push_back(ObjContInfo::CIF_DDL_ITEM_PARENT_NAME);
    values.push_back(_id);   

    tbl->SetFlags(ObjContInfo::CIF_DDL_ITEM_CHILD_NAME, ISTable::DT_STRING |
      ISTable::CASE_INSENSE);

    tbl->Search(listOut, values, ColumnNamesCh);
    if (listOut.empty())
        return;
  
    len = 2*tbl->GetNumRows()*sizeof(int)+1;
    parentList = (int *) malloc(len);

    num1 = listOut.size();
    for (i=0;i<num1; i++) {
	 parentList[2*pCount] = listOut[i];
	 parentList[2*pCount + 1] = generationNum;
	 pCount++;
    }
    while (totalCount<pCount) {
	 values.clear();
	 cell.clear();
	 //tbl->GetCell(cell, ObjContInfo::CIF_DDL_ITEM_PARENT_NAME,
         //  parentList[2*totalCount]);
         cell = (*tbl)(parentList[2*totalCount],
           ObjContInfo::CIF_DDL_ITEM_PARENT_NAME);
	 values.push_back(cell);
	 cell.clear();
	 tbl->Search(listOut, values, ColumnNamesCh);
	 if (!listOut.empty()) {
		generationNum = parentList[2*totalCount + 1] + 1;
		num = listOut.size();
		for (i=0;i<num; i++) {
		  parentList[2*pCount] = listOut[i];
		  parentList[2*pCount + 1] = generationNum;
		  pCount++;
		}
	 }
	 totalCount++;
    }

    family_member familyMember;
    for (i = 0; i < totalCount; i++)
    {
	 //tbl->GetCell(cell, ObjContInfo::CIF_DDL_ITEM_PARENT_NAME,
         //  parentList[2*i]);
         cell = (*tbl)(parentList[2*i],
           ObjContInfo::CIF_DDL_ITEM_PARENT_NAME);
         familyMember.name = cell;
	 cell.clear();
         familyMember.generation = parentList[2 * i + 1];

         decendency.push_back(familyMember);
    }
  
    if (parentList) free(parentList);
}


void ItemObjCont::BuildItems(vector<vector<string> >& combo,
  const unsigned int configIndex)
{
    if (_id.empty() || _blockName.empty())
        return;

    ObjCont::BuildItems(combo, configIndex);

    if (!combo[0].empty())
        return;
    else
    {
        for (unsigned int i = 0; i < _decendency.size(); i++)
        {
            ObjCont::BuildItems(combo, configIndex, _decendency[i].name);
            if (!combo[0].empty())
                break;
        }
        return;
    }
}


DictObjCont::DictObjCont(FileNavigator& fileNav, const string& blockName)
  : ObjCont(fileNav, *(new DicFile), blockName, blockName, RcsbDict)
{
    Init();
}


DictObjCont::DictObjCont(FileNavigator& fileNav, DicFile& dicFile,
  const string& blockName) : ObjCont(fileNav, dicFile, blockName, blockName,
  RcsbDict)
{
    Init();
}


DictObjCont::~DictObjCont()
{
    _verbose = false;
    // VLAD - IMPLEMENT - PROPERLY IMPLEMENT THIS DESTRUCTOR
}


void DictObjCont::Build()
{

    ObjCont::Build();

    unsigned int index =
      RcsbDict.GetItemIndex(ObjContInfo::CIF_DDL_CATEGORY_CATEGORY,
      ObjContInfo::CIF_DDL_ITEM_ID);
    sort(_itemsStore[index][0].begin(), _itemsStore[index][0].end());

    index =
      RcsbDict.GetItemIndex(ObjContInfo::CIF_DDL_CATEGORY_SUB_CATEGORY,
      ObjContInfo::CIF_DDL_ITEM_ID);
    sort(_itemsStore[index][0].begin(), _itemsStore[index][0].end());

    index =
      RcsbDict.GetItemIndex(ObjContInfo::CIF_DDL_CATEGORY_ITEM,
      ObjContInfo::CIF_DDL_ITEM_NAME);
    sort(_itemsStore[index][0].begin(), _itemsStore[index][0].end());

}


int DictObjCont::Write()
{
    ItemObjCont* ilo;
    ObjCont* slo;
    ObjCont* clo;

    pair<unsigned int, unsigned int> indexPair =
      RcsbDict.GetItemIndices(ObjContInfo::CIF_DDL_CATEGORY_ITEM,
      ObjContInfo::CIF_DDL_ITEM_NAME);

    vector<uWord> itemsIndices;

    for (unsigned int i = 0; i < _itemsStore[indexPair.first]\
      [indexPair.second].size(); i++)
    {
        ilo = new ItemObjCont(_fileNav, _dicFile, _blockName,
          _itemsStore[indexPair.first][indexPair.second][i]);
        ilo->Build();
        if (_verbose) ilo->Print();
        itemsIndices.push_back(ilo->Write());
        delete ilo;
    }

    indexPair = RcsbDict.GetItemIndices(ObjContInfo::\
      CIF_DDL_CATEGORY_SUB_CATEGORY, ObjContInfo::CIF_DDL_ITEM_ID);

    vector<uWord> subcategoriesIndices;

    for (unsigned int i = 0; i < _itemsStore[indexPair.first]\
      [indexPair.second].size(); i++)
    {
        slo = new ObjCont(_fileNav, _dicFile, _blockName,
          _itemsStore[indexPair.first][indexPair.second][i], RcsbSubcat);
        slo->Build();
        if (_verbose) slo->Print();
        subcategoriesIndices.push_back(slo->Write());
        delete slo;
    }

    indexPair = RcsbDict.GetItemIndices(ObjContInfo::\
      CIF_DDL_CATEGORY_CATEGORY, ObjContInfo::CIF_DDL_ITEM_ID);

    vector<uWord> categoriesIndices;

    for (unsigned int i = 0; i < _itemsStore[indexPair.first]\
      [indexPair.second].size(); i++)
    {
        clo = new ObjCont(_fileNav, _dicFile, _blockName,
          _itemsStore[indexPair.first][indexPair.second][i], RcsbCat);
        clo->Build();
        if (_verbose) clo->Print();
        categoriesIndices.push_back(clo->Write());
        delete clo;
    }

    _index.push_back(WriteContLocations(itemsIndices));
    _index.push_back(WriteContLocations(subcategoriesIndices));
    _index.push_back(WriteContLocations(categoriesIndices));

    return(ObjCont::Write());
}


void DictObjCont::Read(uWord which)
{

    // HARDCODED - VLAD - FORMAT DEPENDENT
    ObjCont::Read(which, 3);

    // HARDCODED - VLAD - FORMAT DEPENDENT
    BuildContainers(0, ObjContInfo::CIF_DDL_CATEGORY_ITEM,
      ObjContInfo::CIF_DDL_ITEM_NAME, _items);
    BuildContainers(1, ObjContInfo::CIF_DDL_CATEGORY_SUB_CATEGORY,
      ObjContInfo::CIF_DDL_ITEM_ID, _subcategories);
    BuildContainers(2, ObjContInfo::CIF_DDL_CATEGORY_CATEGORY,
      ObjContInfo::CIF_DDL_ITEM_ID, _categories);

}


ObjCont& DictObjCont::GetObjCont(const string& contName,
  const ObjContInfo& objContInfo)
{

    if (objContInfo._objContInfoDescr == "item")
    {
        return(GetContainers(contName, _items, objContInfo));
    }
    else if (objContInfo._objContInfoDescr == "subcategory")
    {
        return(GetContainers(contName, _subcategories, objContInfo));
    }
    else if (objContInfo._objContInfoDescr == "category")
    {
        return(GetContainers(contName, _categories, objContInfo));
    }
    else
    {
        throw NotFoundException("Container type not found",
          "DictObjCont::GetObjCont");
    }

    throw NotFoundException("Container type not found",
      "DictObjCont::GetObjCont");

}


void DictObjCont::Print()
{

    PrintContainers(ObjContInfo::CIF_DDL_CATEGORY_CATEGORY,
      ObjContInfo::CIF_DDL_ITEM_ID, RcsbCat);
    PrintContainers(ObjContInfo::CIF_DDL_CATEGORY_SUB_CATEGORY,
      ObjContInfo::CIF_DDL_ITEM_ID, RcsbSubcat);
    PrintContainers(ObjContInfo::CIF_DDL_CATEGORY_ITEM,
      ObjContInfo::CIF_DDL_ITEM_NAME, RcsbItem);

    ObjCont::Print();

}


void DictObjCont::PrintContainers(const string& catName,
  const string& itemName, const ObjContInfo& objContInfo)
{

    pair<unsigned int, unsigned int> indexPair =
      RcsbDict.GetItemIndices(catName, itemName);

    for (unsigned int i = 0; i < _itemsStore[indexPair.first]\
      [indexPair.second].size(); ++i)
    {
        ObjCont& objCont = GetObjCont(_itemsStore[indexPair.first]\
          [indexPair.second][i], objContInfo);
        try
        {
            objCont.Print();
#ifdef VLAD_DELETED
            // This is wrong. Why would we delete upon printing?
            delete(objContP);
#endif
        }
        catch (NotFoundException)
        {
            cout << "Failed to load " << objContInfo._objContInfoDescr <<
              " " << _itemsStore[indexPair.first][indexPair.second][i] <<
              endl;
        }
    }

}


void DictObjCont::Init()
{

}


int DictObjCont::WriteContLocations(const vector<uWord>& indices)
{
    if (indices.empty())
        return(-1);

    Word i;

    int err = _fileNav.WriteUWords(indices, i);

    if (err == NO_ERROR)
    {
        // If no error, return the index that the list of idices was written to
        return(i);
    }
    else
    {
        // If error, return the error code, that is less than zero
        return(err);
    }
}


int DictObjCont::BuildContainers(unsigned int index,
  const string& catName, const string& itemName,
  mapped_ptr_vector<ObjCont>& containers)
{
    if (_index[index] == 0)
        return BUILD_LIST_ERROR;

    pair<unsigned int, unsigned int> indexPair =
      RcsbDict.GetItemIndices(catName, itemName);

    int err;

    vector<uWord> ddlContIndices;
    _fileNav.GetUWords(ddlContIndices, _index[index], err);

    containers.push_back(_itemsStore[indexPair.first][indexPair.second],
      ddlContIndices);
 
    return NO_ERROR;
}


void DictObjCont::BuildItems(vector<vector<string> >& combo,
  const unsigned int configIndex)
{

    ISTable* isTableP = NULL;
    vector<string> columnNames;

    if (_blockName.empty())
    {
        return;
    }

    Block& block = _dicFile.GetBlock(_blockName);

    isTableP = block.GetTablePtr(RcsbDict._cats[configIndex].catName);
    if (isTableP == NULL)
    {
        return;
    }

    for (unsigned int ind = 0; ind < RcsbDict._cats[configIndex].\
      items.size(); ++ind)
    {
        columnNames.push_back(RcsbDict._cats[configIndex].\
          items[ind].itemName);
    }
  
    for (unsigned int ind = 0; ind < RcsbDict._cats[configIndex].\
      items.size(); ++ind)
    {
        isTableP->GetColumn(combo[ind], columnNames[ind]);
    }

}


ObjCont& DictObjCont::GetContainers(const string& contName,
  mapped_ptr_vector<ObjCont>& containers, const ObjContInfo& objContInfo)
{
    unsigned int index = containers.find(contName);
    if (index == containers.size())
    {
        throw NotFoundException("Container not found",
          "DictObjCont::GetContainers");
    }

#ifdef VLAD_MODIFY
    if (containers[index] != NULL)
        return containers[index];
#else
    if (containers.is_read(contName))
        return(containers[index]);
#endif

#ifdef VLAD_MODIFY
    containers[index] = new ObjCont(_fileNav, _dicFile, _blockName, contName,
      objContInfo);
#else
    containers.set(new ObjCont(_fileNav, _dicFile, _blockName,
      contName, objContInfo));
#endif
    containers.read(contName);

    return(containers[index]);
}

